#ifndef DIG_FILE_SNARK_REC_H
#define DIG_FILE_SNARK_REC_H

//#include "DIGFile.h"
#include <DIGFile/DIGFile.h>

///////////////////////////////////////////////
// DIGFileSnarkRec main header
///////////////////////////////////////////////

class RecFileMH
{
public:
  DIGString     Title;
  unsigned int  Dimensions;
  double        Sampling;
  DIGString     Comment;
};

///////////////////////////////////////////////
// DIGFileSnarkRec phantom header
///////////////////////////////////////////////

class RecFilePhanomH
{
public:
  DIGString     Name;
  DIGString     Comment;
};

///////////////////////////////////////////////
// DIGFileSnarkRec Reconstruction Set header
///////////////////////////////////////////////

class RecFileRecSetH
{
public:
  DIGString   Name;
  DIGString   AlgName;
  DIGString   Comment;
};

///////////////////////////////////////////////
// DIGFileSnarkRec Reconstruction header
///////////////////////////////////////////////

class RecFileRecH
{
public:
  unsigned int  Count;
  DIGString     Comment;
};

///////////////////////////////////////////////
// DIGFileSnarkRec
///////////////////////////////////////////////

class DIGFileSnarkRec: protected DIGFile {

protected:
  static const char*  TypeStr;
  static const char*  SchemaStr;
  bool                PhantomPresent;
  bool                RecPressent;

public:

	DIGFileSnarkRec();
	~DIGFileSnarkRec();

  //
  // Writing
  //

  // Open for writing

  int Open(const char* pFileName, RecFileMH* pMainHeader);
  int Open(
    const char* pFileName,      // File Name
    const char* pProjName,      // Projection Name
    unsigned int pNoOfElements, // NELEM
    double pSampling,           // PIXSIZ
    const char* pComments       // Comments
  );

  // Updating main header after opening file to write

  int SetProjName(const char* pProjName) {
    return DIGFile::SetTitle(pProjName);
  }

  int SetComment(const char* pComment) {
    return DIGFile::SetComment(pComment);
  }

  // Appending phantom

  int AppendPhanom(RecFilePhanomH* pPhanomHeader, double* pData);
  int AppendPhantom(const char* pPhanomName, const char* pComment, double* pData);

  // Appending reconstruction set

  int AppendRecSet(RecFileRecSetH* pRecHeader);
  int AppendRecSet(const char* pAlgName, const char* pName, const char* pComment);

  // Appending reconstruction

  int AppendRec(RecFileRecH* pRecHeader, double* pData);
  int AppendRec(unsigned int pCount, const char* pComment, double* pData);

  //
  // Reading
  //

  // Open for reading

  int Open(const char* pFileName);

  // Reading file parameters

  int GetPhantomPresent(bool* pPhantomPresent) {
    *pPhantomPresent = PhantomPresent;
    return 0;
  }

  int GetNoOfPixels(unsigned int* pNoOfPixels) {
    return DIGFile::GetArrayNoOfItems(pNoOfPixels);
  }

  int GetNoOfRecSets(unsigned int* pNoOfRecSets) {
    int ret;
    unsigned int NoOfRecSets;

    if((ret = DIGFile::GetNoOfArraySets(&NoOfRecSets)) != 0) {
      return ret; // error
    }

    if(PhantomPresent) {
      *pNoOfRecSets = NoOfRecSets - 1;
    }
    else {
      *pNoOfRecSets = NoOfRecSets;
    }

    return 0;
  }

  int GetNoOfRecs(unsigned int* pNoOfRec) {
    return GetNoOfArrays(pNoOfRec);
  }

  // file navigation

  int SelectRecSet(unsigned int pRecSetNo) {
    if(PhantomPresent) {
      return SelectArraySet(pRecSetNo + 1);
    }
    else {
      return SelectArraySet(pRecSetNo);
    }
  }

  int SelectRec(unsigned int pRecNo) {
    return SelectArray(pRecNo);
  }

  // Reading main header

  int GetMainHeader(RecFileMH* pMainHeader);

  int GetProjName(const char** pProjName) {
    return DIGFile::GetTitle(pProjName);
  }

  int GetEndian(DIGEndian* pEndian) {
    return DIGFile::GetEndian(pEndian);
  }

  int GetDimensions(unsigned int* pDimensions) {
    const DIGDimensions* Dimensions;
    int ret;

    if((ret = DIGFile::GetDimensions(&Dimensions)) != 0) {
      return ret;
    }

    *pDimensions = Dimensions->x;

    return 0;
  }

  int GetSampling(double* pSampling) {
    const DIGSampling* Sampling;
    int ret;

    if((ret = DIGFile::GetSamplingX(&Sampling)) != 0) {
      return ret;
    }

    *pSampling = Sampling->x;

    return 0;
  }

  int GetComment(const char** pComment) {
    return DIGFile::GetComment(pComment);
  }

  // Reading phantom header

  int GetPhantomHeader(RecFilePhanomH* pPhantomHeader);

  int GetPhantomName(const char** pPhantomName);
  int GetPhantomComment(const char** pPhantomComment);

  // Reading phantom data

  int GetPhantomData(void* pPhantomData);

  // Reading reconstruction set header

  int GetRecSetHeader(RecFileRecSetH* pRecSetHeader);

  int GetRecSetName(const char** pRecName); // title
  int GetRecSetAlgName(const char** pAlgName); // type
  int GetRecSetComment(const char** pComment);

  // Reading reconstruction header

  int GetRecHeader(RecFileRecH* pRecHeader);

  int GetRecCount(unsigned int* pCount);
  int GetRecComment(const char** pComment);

  // Reading reconstruction data

  int GetRecData(void* pRecData) {
    return DIGFile::GetArrayData(pRecData);
  }

  ///////////////////////////////////////////////////

  int Close() {
    return DIGFile::Close();
  }

};

#endif // DIG_FILE_SNARK_REC_H
