/*
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/back.cpp $
 $LastChangedRevision: 85 $
 $Date: 2014-07-02 16:07:08 -0400 (Wed, 02 Jul 2014) $
 $Author: agulati $
 ***********************************************************

 COMBINED BACK-PROJECTION RECONSTRUCTION ALGORITHM.

 THIS SUBROUTINE COMPUTES A BACK-PROJECTION RECONSTRUCTION
 USING EITHER THE EQUATIONS OF THE LINES AS GIVEN BY RAY OR WRAY
 OR BY INTERPOLATING BETWEEN THE GIVEN PROJECTION DATA TO
 APPROXIMATE THE RADON TRANSFORM AT A PARTICULAR POINT.  THE
 FORMER IS SPECIFIED BY THE 'DISCRETE' PARAMETER, THE LATTER BY
 'CONTINUOUS'.

 IN THE DISCRETE CASE, THE USER HAS THE ADDITIONAL OPTIONS OF
 SPECIFYING IF THE RAYSUM IS TO BE DIVIDED BY 'SNORM' BEFORE
 BEING BACK-PROJECTED ('DISTRIBUTED') AND IF PIXELS WHICH ARE
 INTERCEPTED BY A ZERO OR NEGATIVE RAY SHOULD BE REMOVED FROM
 THE COMPUTATIONS ('ZEROIZING').

 IN THE CONTINUOUS CASE, THE DESIRED INTERPOLATION METHOD MUST
 BE SPECIFIED.  WARNING- THE CONTINUOUS CASE WEIGHTS PROJECTION
 "NP" BY "(THETA(NP+1)-THETA(NP-1))/2.0".  IN ORDER TO ACCOMPLISH
 THIS IT IS ASSUMED THAT THE PROJECTIONS ARE ARRANGED IN INCREASING
 ORDER ON THETA AND THAT THETA(PRJNUM)-THETA(1) IS LESS THAN
 PI IN THE PARALLEL CASE OR TWOPI IN THE DIVERGENT CASE.

 THE RESULT OF THE BACK-PROJECTION WILL BE NORMALIZIED TO
 THE CORRECT AVERAGE DENSITY IF EITHER 'ADDITIVE' OR 'MULTIPLICATIV
 HAS BEEN SPECIFIED.  THE FORMER PERFORMS THE NORMALIZATION BY
 THE ADDITION OF A CONSTANT TO THE ENTIRE PICTURE, THE LATTER BY
 MULTIPLICATION.
 */

#include <cstdio>
#include <cmath>

#include "blkdta.h"
#include "geom.h"
#include "spctrm.h"
#include "fourie.h"
#include "consts.h"
#include "uiod.h"

#include "bckprj.h"
#include "anglst.h"
#include "ray.h"
#include "wray.h"

#include "projfile.h"
#include "infile.h"

#include "back.h"

BOOLEAN back_class::Run(REAL* recon, INTEGER* list, REAL* weight, INTEGER iter)
{
	REAL neginf, zflag;
	INTEGER word;
	REAL* g = NULL;       //wei 3/2005
	BOOLEAN disc, zeros, dist, add, mult;
	BOOLEAN eol;
	static const INTEGER hcont = CHAR2INT('c', 'o', 'n', 't');
	static const INTEGER hdisc = CHAR2INT('d', 'i', 's', 'c');
	static const INTEGER hdist = CHAR2INT('d', 'i', 's', 't');
	static const INTEGER hzero = CHAR2INT('z', 'e', 'r', 'o');
	static const INTEGER haddi = CHAR2INT('a', 'd', 'd', 'i');
	static const INTEGER hmult = CHAR2INT('m', 'u', 'l', 't');

	// modified calls to getwrd() to use the 4-parameter version. Lajos, Dec 13, 2004
	static const INTEGER back_codes[6] =
	{
			CHAR2INT('c', 'o', 'n', 't'),
			CHAR2INT('d', 'i', 's', 'c'),
			CHAR2INT('z', 'e', 'r', 'o'),
			CHAR2INT('d', 'i', 's', 't'),
			CHAR2INT('a', 'd', 'd', 'i'),
			CHAR2INT('m', 'u', 'l', 't') };

	INTEGER nsize;
	REAL thmod;
	REAL thfac;
	REAL theta;
	REAL sinth;
	REAL costh;
	REAL theta0;
	INTEGER np;
	REAL theta1;
	INTEGER np2;
	REAL theta2;
	REAL sinth2;
	REAL costh2;
	REAL w;
	INTEGER nr;
	INTEGER numb;
	REAL snorm;
	INTEGER nb;
	INTEGER k;
	REAL amt;
	REAL sum;
	INTEGER knt;
	REAL corr;

	if (iter > 1)
		return FALSE;
	neginf = -Consts.infin;
	zflag = neginf + neginf;
	dist = FALSE;
	zeros = FALSE;
	add = FALSE;
	mult = FALSE;

	// PROCESS CONTROL CARD

	word = InFile.getwrd(TRUE, &eol, back_codes, 2);
	if (word != hcont)
	{
		if (word != hdisc)
		{
			fprintf(output, "\n **** neither continuous or discrete specified");
			return TRUE;
		}

		disc = TRUE;

		word = InFile.getwrd(FALSE, &eol, &(back_codes[2]), 4);

		if (word == hzero)
		{
			zeros = TRUE;
			word = InFile.getwrd(FALSE, &eol, &(back_codes[3]), 3);
		}

		if (word == hdist)
		{
			dist = TRUE;
			word = InFile.getwrd(FALSE, &eol, &(back_codes[4]), 2);
		}

	}
	else
	{
		disc = FALSE;
		Fourie.interp = InFile.getint(FALSE, &eol);
		if ((eol) || (Fourie.interp < -1) || (Fourie.interp > 6))
		{
			// ERROR CONDITIONS

			fprintf(output, "\n **** invalid interpolation method");

			return TRUE;
		}

		word = InFile.getwrd(FALSE, &eol, &(back_codes[4]), 2);
	}

	if (!eol)
	{
		if (word == haddi)
			add = TRUE;
		if (word == hmult)
			mult = TRUE;
	}

	if (disc)
	{
		fprintf(output, "\n          discrete back-projection");
	}
	else
	{

		fprintf(output, "\n          continuous back-projection");

		if (Fourie.interp == -1)
		{
			fprintf(output, "\n          modified cubic spline interpolation");
		}

		if (Fourie.interp == 0)
		{
			fprintf(output, "\n          band limiting sinc interpolation");
		}

		if (Fourie.interp > 0)
		{
			fprintf(output, "\n          %2i order lagrange interpolation",
					Fourie.interp);
		}
	}

	if (dist)
	{
		fprintf(output, "\n          distributed back-projection");
	}

	if (zeros)
	{
		fprintf(output, "\n          zeroizing is in effect");
	}

	if (add)
	{
		fprintf(output, "\n          additive normalization");
	}

	if (mult)
	{
		fprintf(output, "\n          multiplicative normalization");
	}

	if (!disc)
	{

		// CONTINUOUS CASE

		nsize = GeoPar.snrays + 2 * Fourie.interp;
		if (Fourie.interp == -1)
			nsize = GeoPar.snrays + 6;

		g = new REAL[nsize];

		thmod = Consts.pi;
		thfac = 2.0;
		if (!GeoPar.par)
		{
			thmod = Consts.twopi;
			thfac = 4.0;
		}
		Anglst.getang(GeoPar.prjnum - 1, &theta, &sinth, &costh);
		theta0 = theta - thmod;

		for (np = 0; np < GeoPar.prjnum; np++)
		{

			ProjFile.ReadProj(np, g, nsize);

			Anglst.getang(np, &theta1, &sinth, &costh);

			np2 = np + 1;

			if (np == (GeoPar.prjnum - 1))
				np2 = 0;

			Anglst.getang(np2, &theta2, &sinth2, &costh2);

			if (np == (GeoPar.prjnum - 1))
				theta2 += thmod;

			w = (theta2 - theta0) / thfac;

			if (w <= 0.0)
			{
				fprintf(output,
						"\n **** projection angles not in increasing order");

				if (g != NULL)
					delete[] g;   //wei 3/2005
				return TRUE;
			}

			Fourie.rinc = GeoPar.pinc;

			if (GeoPar.div)
				Fourie.rinc *= GeoPar.radius / GeoPar.stod;
			if (GeoPar.vri)
				Fourie.rinc *= (REAL) MAX0(fabs(sinth), fabs(costh));
			if (GeoPar.strip)
				w /= Fourie.rinc;

			theta0 = theta1;

			for (nr = 0; nr < nsize; nr++)
			{
				g[nr] *= w;
			}

			bckprj(recon, GeoPar.nelem, g, nsize, sinth, costh,
					GeoPar.pixsiz / Fourie.rinc, Fourie.interp);

		}

	}
	else
	{
		// DISCRETE CASE

		g = new REAL[GeoPar.nrays];

		if (zeros)
		{

			// FLAG PIXELS INTERCEPTED WITH ZERO OR NEGATIVE RAYS

			for (np = 0; np < GeoPar.prjnum; np++)
			{

				ProjFile.ReadProj(np, g, GeoPar.nrays);

				for (nr = GeoPar.fsnray; nr <= GeoPar.lsnray; nr++)
				{ // changed < to <=. lajos Nov 17, 2004
					if (g[nr] <= Consts.zero)
					{

						if (GeoPar.strip)
							ray(np, nr, list, weight, &numb, &snorm);
						if (GeoPar.line)
							wray(np, nr, list, weight, &numb, &snorm);

						if (numb != 0)
						{

							for (nb = 0; nb < numb; nb++)
							{
								k = list[nb];
								recon[k] = zflag;
							}
						}
					}
				}
			}

			// PERFORM DISCRETE BACK-PROJECTION
		}

		for (np = 0; np < GeoPar.prjnum; np++)
		{

			ProjFile.ReadProj(np, g, GeoPar.nrays);
			for (nr = GeoPar.fsnray; nr <= GeoPar.lsnray; nr++)
			{ // changed < to <=. lajos Nov 17, 2004

				if (GeoPar.strip)
					ray(np, nr, list, weight, &numb, &snorm);
				if (GeoPar.line)
					wray(np, nr, list, weight, &numb, &snorm);

				if (numb != 0)
				{
					amt = g[nr];
					if (dist)
					{
						if (zeros)
						{

							// CORRECT SNORM BY REMOVING FLAGGED PIXELS
							snorm = 0.0;
							for (nb = 0; nb < numb; nb++)
							{
								k = list[nb];
								if (recon[k] > neginf)
									snorm += weight[nb] * weight[nb];
							}
						}
						amt /= snorm;
					}
					for (nb = 0; nb < numb; nb++)
					{
						k = list[nb];
						recon[k] += amt * weight[nb];
					}
				}
			}

		}
	}

	// PERFORM NORMALIZATION

	if (!(add || mult))
	{
		if (!zeros)
		{
			if (g != NULL)
				delete[] g;   //wei 3/2005
			return FALSE;
		}
		// NO NORMALIZATION REQUESTED - JUST CLEAR FLAGGED PIXELS

		for (k = 0; k < GeoPar.area; k++)
		{
			if (recon[k] <= neginf)
				recon[k] = 0.0;
		}

		if (g != NULL)
			delete[] g;   //wei 3/2005
		return FALSE;

		// COMPUTE TOTAL DENSITY AND NUMBER OF UNFLAGGED PIXELS
	}

	sum = 0.0;
	knt = 0;

	for (k = 0; k < GeoPar.area; k++)
	{

		if (recon[k] > neginf)
		{
			sum += recon[k];
			knt++;
		}

	}

	if (!mult)
	{

		// ADDITIVE CORRECTION TO AVERAGE DENSITY

		corr = GeoPar.aveden - sum / ((REAL) (knt));

		for (k = 0; k < GeoPar.area; k++)
		{
			recon[k] += corr;
			if (recon[k] <= neginf)
				recon[k] = 0.0;
		}
		if (g != NULL)
			delete[] g;   //wei 3/2005
		return FALSE;

		// MULTIPLICATIVE CORRECTION TO AVERAGE DENSITY
	}

	corr = ((REAL) (knt)) * GeoPar.aveden / sum;

	for (k = 0; k < GeoPar.area; k++)
	{
		if (recon[k] <= neginf)
			recon[k] = 0.0;
		recon[k] *= corr;
	}
	if (g != NULL)
		delete[] g;   //wei 3/2005
	return FALSE;
}

