/*
***********************************************************
$SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
$HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/blob_functions.cpp $
$LastChangedRevision: 45 $
$Date: 2014-06-12 19:31:23 -0400 (Thu, 12 Jun 2014) $
$Author: olangthaler $
***********************************************************

 There are three subroutines:

 1. dc  computes the DC term of the blob defined by the function kaiser 
    (i.e., Robert Lewitt's original definition of a blob) blob function at 
    distance "dist",

 2. radial computes the normalized blob function at distance "dist", and 

 3. lintegral computes the line integral of a blob at distance "dist".

*/

#include <cstdio>
#include <cmath>

#include "consts.h"
#include "blkdta.h"
#include "blob.h"

REAL kaiser_proj(REAL s);
INTEGER kaiser(REAL* temp, REAL dist);
static REAL bessi3(REAL shape);
static REAL bessi2(REAL shape);

REAL Blob_class::dc()
{   
  REAL ans;
  REAL r = Blob.support;
  REAL shape = Blob.shape;
  //REAL fshape = fabs(shape); shape is supposted to be non-negative

  ans = 2 * Consts.pi * r * r * bessi3(shape) / (shape * bessi2(shape));
  //tracer
  //fprintf(output,"\ndc = %f  ", ans); 
  //tracer   
  return ans;
}


REAL Blob_class::radial(REAL dist)
{   
  REAL ans;
  REAL dc = Blob.dc(); 
 
  kaiser(&ans, dist);
  // now comes the normalization
  ans = ans * Consts.sqrt3 / 2.0 * Blob.delta * Blob.delta / dc;
 //tracer
  //fprintf(output,"\nblob factor = %f  ", sqrt3 / 2.0 * Blob.delta * Blob.delta / dc); 
  //tracer   
   return ans;
}

REAL Blob_class::lintegral(REAL dist)
{   
  REAL ans;
  REAL dc = Blob.dc(); 
  
  ans = kaiser_proj(dist);
  ans = ans * Consts.sqrt3 / 2.0 * Blob.delta * Blob.delta / dc;  //normalisation

  return ans;
}


// Function to compute the Bessel 0

static REAL bessi0(REAL x)
{
  REAL y,ax,ans;

  ax = fabs(x);
  if(ax < 3.75) {
    y = (x * x) / (3.75 * 3.75);
    ans = 1.0 + y * (3.5156229 + y * (3.0899424 + y * (1.2067492
        + y * (0.2659732 + y * (0.360768e-1 + y * 0.45813e-2)))));
  }
  else {
    y = 3.75 / ax;
    ans = (exp(ax) / sqrt(ax)) * (0.39894228 + y * (0.1328592e-1
        + y * (0.225319e-2 + y * (-0.157565e-2 + y * (0.916281e-2
        + y * (-0.2057706e-1 + y * (0.2635537e-1 + y * (-0.1647633e-1
        + y * 0.392377e-2))))))));
  }
  return ans;
}

// Function to compute the Bessel 1
static REAL bessi1(REAL x)
{
  REAL ax, ans, y;

  ax = fabs(x);
  if(ax < 3.75) {
    y = x / 3.75;
    y *= y;
    ans = ax * (0.5 + y * (0.87890594 + y * (0.51498869 + y * (0.15084934
        + y * (0.2658733e-1 + y * (0.301532e-2 + y * 0.32411e-3))))));
  }
  else {
    y = 3.75 / ax;
    ans = 0.2282967e-1 + y * (-0.2895312e-1 + y * (0.1787654e-1
        - y * 0.420059e-2));
    ans = 0.39894228 + y * (-0.3988024e-1 + y * (-0.362018e-2
        + y * (0.163801e-2 + y * (-0.1031555e-1 + y * ans))));
    ans *= (exp(ax) / sqrt(ax));
  }

  return (x < 0.0)? -ans : ans;
}


// Function to compute the Bessel 2

static REAL bessi2(REAL x)
{
  REAL ans;

  if(x <= Consts.zero) {
    ans = 0.0;
  }
  else {
    ans = bessi0(x) - (2.0 / x) * bessi1(x);
  }

  return ans;
}

// Function to compute the Bessel 3
static REAL bessi3(REAL x)
{
  REAL ans;

  if(x <= Consts.zero) {
    ans = 0.0;
  }
  else {
    ans = bessi1(x) - (4.0 / x) * bessi2(x);
  }
  return ans;
}
/* Function to compute the value of the Spatial Kaiser-Bessel Function */
/*
 * dist stands for the radial distance from the blob center (variable)
 * CONSTANTS:
 * Blob.support stands for the radius
 * Blob.shape for the decay (shape of the blob)
 *
 */

INTEGER kaiser(REAL* ans, REAL dist)
{
  REAL rda,rdas,arg,w;

  rda = dist / Blob.support;
  rdas = rda * rda;
  if(rdas <= 1.0) {
    arg = Blob.shape * sqrt(1.0 - rdas);

    if(fabs(Blob.shape) < Consts.zero) {
      // if(Blob.shape==0.0)
      w = (1.0 - rdas) * (1.0 - rdas);
    }
    else {
      w = (1.0 - rdas) * bessi2(arg) / bessi2(Blob.shape);
      //fprintf(output,"\ny=%f",*y);
    }
  }
  else {
    w = 0.0;
  }

  *ans = w;

  return 1;
}


// Function to compute the value of the Spatial Kaiser-Bessel Function
//
//      Value of line integral through Kaiser-Bessel radial function
//      (n >=2 dimensions) at distance s from center of function.
// Blob.support stands for the radius
// Blob.shape for the decay (shape of the blob)

REAL kaiser_proj(REAL s)
{
  REAL sda, sdas, w, arg, p;

  sda = s / Blob.support;
  sdas = sda * sda;
  w = 1.0 - sdas;
  if(w > 1.0e-10) {
    arg = Blob.shape * sqrt(w);

    if(Blob.shape == 0.0) {
      p = 2.0 * Blob.support * w * w * sqrt(w) * (8.0 / 15.0);
    }
    else {
      p = (2.0 * Blob.support/Blob.shape) * w *\
        ((3.0/(arg * arg) + 1.0) * sinh(arg) - (3.0/arg) * cosh(arg) ) / bessi2(Blob.shape);
    }
  }   
  else {
    p = 0.0;
  }

  return p;
}


