/*
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                               *
 *                         S N A R K   1 4                       *
 *                                                               *
 *                A PICTURE RECONSTRUCTION PROGRAM               *
 *                                                               *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

  supersnk.c,v 1.16 2011/02/21 00:45:48 jklukowska Exp

 Previously part of SuperSNARK*/

//#define SS_DEBUG  //bug 269, Joanna Klukowska, converted to a variable below

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>

#include "experimenter.h"

#include "get_phantom.h"
#include "create_phantom.h"
#include "projection.h"
#include "recon.h"
#include "analysis.h"
#include "errorc.h"
#include "analyze.h"
#include "analyze2.h"
#include "get_iter_flag.h"
//#include "get_seed.h"
#include "check_if_iroi.h"
#include "snark.h"

/* --------------------------- supersnk.c -------------------------------
 *
 * This function is the driver program to SNARK. It controls phantom
 * selection and feeds SNARK the necessary commands to generate the phantoms
 * plus their projection data, to implement the desired reconstruciton
 * algorithms on these data, and to evaluate the reconstructed images.
 * This function also calls the necessary routines to compute the
 * desired figures of merit and their statistical significance.
 *
 * INPUTS:
 * nexp - number of experiments.
 * expflg - controls the phantom-file set used in each experiment as
 * follows: if expflg=0 the same set of phantom files is available
 * for use in both experiments. If expflg=1 the phantom files used
 * in the first experiment are excluded in the second experiment.
 * navel, nelem,pixsz - modifiers described by Set 4 of the SNARK CREATE
 * input sequence (see the SNARK manual for details).
 * nphan - two-element array: nphan[i-1]= number of phantom files used for
 * experiment i (i=1,2).
 * nrun - two-element array: nrun[i-1]= number of runs for experiment i
 * (i=1,2).
 * ensflname - name of ensemble data file.
 * projflname - name of file containing information about the projection
 * data.
 * recflname - name of file containing information about the reconstruction
 * algorithms to be implemented.
 * anaflname - name of file containing information about the algorithms
 * being compared.
 *
 * OUTPUTS:
 * none
 * -------------------------------------------------------------------- */


void supersnk(
        int nexp,
        int expflg,
        int nelem,
        float pixsz,
        int navel,
        int* nphan,
        int* nrun,
        char* ensflname,
        char* projflname,
        char* recflname,
        char* anaflname,
        char* progName,
        long seed
        )
//int nexp,expflg,nelem,navel,*nphan,*nrun;
//float pixsz;
//char *ensflname,*projflname,*recflname,*anaflname;
{

    //jk 01/01/08
    //changed pairsarray[MAXNSTR] to *pairsarray
    //changed structure[MAXNSTR][MAXLINESIZE] to **structure
    //to allow unlimited number of structures
    //the file will need to be scanned firest to
    //allocate the array of right size
    int *pairsarray;
    char **structure;
    int i, j, run, szens, nstr, npairs, tot_PAIRS;
    int iroi_used; //set to 0 or 1 by check_if_iroi
    //it is used by create_phantom() to verify that density of one out
    //of each paired  structures is zero
    char namelst[MAXENSEMSIZE][MAXLINESIZE],
            phanflname[MAXLINESIZE],
            currentlst[MAXENSEMSIZE][MAXLINESIZE], fomflname[MAXLINESIZE],
            iter_flag[51], ifname[64], ofname[64], recfilname[64], enums[8], rnums[8], command[256],pipcommand[256];

    FILE *fomfl;

    char * mode;
    int SS_DEBUG = 0;
    mode = getenv("SNARK_EXP");
    if (mode != NULL && strcasecmp(mode, "DEBUG") == 0)
    {
        SS_DEBUG = 1;
        printf("Running in DEBUG mode: extra files will be saved.\n");
    }

    /* get list of phantom filenames and iteration flag for EVALUATE command */
    read_ensemble(ensflname, namelst, &szens);
    strncpy(iter_flag, "10000000000000000000000000000000000000000000000000\0",51);
    getiterflag(anaflname, iter_flag);

    strcat(pipcommand,"tee snark_e.in | (");
    strcat(pipcommand,progName);
    strcat(pipcommand," > snark_e.out)"); 
    for (i = 0; i < nexp; i++) {
        fprintf(stdout, "experiment %d\n", i + 1); // bug 187 - progress indicator - swr - 11/11/05
        /* sprintf(fomflname,"%s.%d",fomflname,i+1); */ /* bug: concats. exp.# onto prev. name ('fomfil.1.2.3....') */
        sprintf(fomflname, "fomfil.%d", i + 1); /* fix: concats. exp.# onto 'fomfil' only (JD 8/8/05) */
        tot_PAIRS = 0; /* number of phantoms with PAIRed structures */
	
        for (run = 0; run < nrun[i]; run++) {
            fprintf(stdout, "run %d\n", run + 1); // bug 187 - progress indicator - swr - 11/11/05
            /* beginning of inner for loop */


            if(SS_DEBUG) {
            //fprintf(stdout, "------------------------------");

                enums[0] = 0;
                sprintf(enums, "%02d", i);

                rnums[0] = 0;
                sprintf(rnums, "%02d", run);

                ifname[0] = 0;
                //strcat(ifname, "snark_e");
                //strcat(ifname, ".in");
                sprintf(ifname, "snark_e%d.in", run);

                ofname[0] = 0;
                //strcat(ofname, "snark_e");
                //strcat(ofname, ".out");
                sprintf(ofname, "snark_e%d.out", run);

                recfilname[0] = '\0';
                sprintf(recfilname, "recfil%d", run);

                if ((pstream = fopen(ifname, "w")) == NULL) {
                    errorc("Could not open", ifname);
                }
            } else {
                // bug 170 - swr - 10/01/05
                if (system("test -x ./bin/snark14UserDefined") || // file exists and executable
                        system("test -s ./bin/snark14UserDefined") || // exists and has non-zero size
                        system("test -f ./bin/snark14UserDefined")) { // exists and is regular (not a directory
        	// bug 298 - Ze Ye 08/13/15
                      if ((pstream = popen("tee snark_e.in | (./bin/snark14 > snark_e.out)", "w")) == NULL) {
                        errorc("Could not run", "snark");
                    }
                } 
			else {
	           if ((pstream = popen("tee snark_e.in | (./bin/snark14UserDefined> snark_e.out )", "w")) == NULL) {
                        errorc("Could not run", "snark");
                    }

                }
            }

            get_phantom(i, run, expflg, nphan, &szens, namelst, currentlst,
                    phanflname);

            check_if_iroi(anaflname, &iroi_used);

            create_phantom(phanflname, seed, navel, nelem, pixsz, &structure,
                    &nstr, &npairs, &pairsarray, iroi_used);
            
            projection(projflname, seed);

            recon(recflname, i + 1, run + 1);

            evaluate(anaflname, &structure, nstr, iter_flag);

            fprintf(pstream, "END \n");
		
            if (SS_DEBUG) {
                fclose(pstream);

                command[0] = 0;
                //strcat(command, "snark14 ");
                strcat(command, progName);
                strcat(command, " ");
                strcat(command, ifname);
                strcat(command, " > ");
                strcat(command, ofname);

                if (system(command) != 0) {
                    fprintf(stdout, "\nERROR running: %s\n", command);
                    exit(1);
                }
                else
                    rename ("recfil", recfilname);

                
            } else {
                pclose(pstream);
            }

            /*    keep count of the number of phantoms */
            if (npairs != 0) {
                tot_PAIRS += 1; /* with PAIRed structures */
            }

            if (run == 0) {
                if ((fomfl = fopen(fomflname, "w")) == NULL) {
                    errorc("in supersnk.c: error opening file ", fomflname);
                }
            } else {
                if ((fomfl = fopen(fomflname, "a")) == NULL) {
                    errorc("in supersnk.c: error opening file ", fomflname);
                }
            }


            analyze(i + 1, run + 1, npairs, pairsarray, anaflname, fomfl);
            fclose(fomfl);

            /* remove snark.out after every run except the very last run */

            if( !SS_DEBUG) {
                if (nexp == 1) {
                    if (run != nrun[0] - 1) {
                        if (remove("snark_e.out") != 0) {
                            errorc("in supersnk: error deleting file ", "snark.out");
                        }
                    }
                } else {
                    if ((i == 0) || ((i == 1) && (run != nrun[1] - 1))) {
                        if (remove("snark_e.out") != 0) {
                            errorc("in supersnk: error deleting file ", "snark_e.out");
                        }
                    }
                }
            }

        } /*    end of inner for loop */

        //jk 01/01/08 free memory
        free(pairsarray);
        for (j = nstr - 1; j >= 0; j--) {
            free((void*) (structure[j]));
        }

        analyze2(i + 1, nrun[i], tot_PAIRS, anaflname);

    } /* end of outer for loop */
}
